#!/usr/bin/env python

from collections import defaultdict

def parse_profile(infile):
  profile=[]
  with open(infile,'r') as freqfile:
    for line in freqfile:
      line=line.strip()
      if line[0]=="#":
        continue
      freq=parse_profile_line(line)
      profile.append(freq)
  return profile

def parse_profile_line(profileline):
  tok=profileline.split()
  freq=dict(d.split(':') for d in tok)
  return defaultdict(lambda:0, {key:float(val) for key,val in freq.items()})

class profile_data:
  def __init__(self,profile):
    self.uniq=[list(col.keys()) for col in profile]
    self.prob=[list(col.values()) for col in profile]
    self.index=[{key:idx for idx,key in enumerate(col.keys())} for col in profile]
    self.cumulative=[0]
    for i in range(len(self.uniq)):
      self.cumulative.append(self.cumulative[i]+len(self.uniq[i]))



#!/usr/bin/env python
import numpy as np
import sys

def encodeInt(seq,profiledata):
  return [ profiledata.index[col][char] for col,char in enumerate(seq) ]

def encodeCategorical(data,profiledata):
#  index=[ profiledata.index[col][d]+profiledata.cumulative[col] for col,d in enumerate(data) ]
  index=[  ]
  for col,d in enumerate(data):
    try:
      index.append( profiledata.index[col][d]+profiledata.cumulative[col] ) 
    except:
      print("I do not know how to encode symbol {} in column {}. This position will remain empty.".format(d,col),file=sys.stderr)
  ret=np.zeros((profiledata.cumulative[-1],),dtype=bool)
  for idx in index:
    ret[idx]=True
  return ret



from datetime import date
import os
import gzip
import shutil
import requests
import time

def download_file(url,local_filename):
  if local_filename is None:
    local_filename = url.split('/')[-1]
  # NOTE the stream=True parameter below
  with requests.get(url, stream=True) as r:
    r.raise_for_status()
    with open(local_filename, 'wb') as f:
      for chunk in r.iter_content(chunk_size=1024*64): 
        if chunk: # filter out keep-alive new chunks
          f.write(chunk)
          f.flush()
  return local_filename

def uniprot(folder=None):
  if folder is None:
    try:
      folder=os.environ['UNIPROTFOLDER']
    except:
      raise RuntimeError("The environment variable UNIPROTFOLDER must be defined")
  if not os.path.isdir(folder):
    os.mkdir(folder)
  
  current_release=os.path.join(folder,"current_release")
  today = date.today()
  folder=os.path.join(folder,today.strftime("%Y%m%d"))
  if not os.path.isdir(folder):
    os.mkdir(folder)
  
  url_uniprot_sprot="https://ftp.expasy.org/databases/uniprot/current_release/knowledgebase/complete/uniprot_sprot.fasta.gz"
  download_file(url_uniprot_sprot,os.path.join(folder,"uniprot_sprot.fasta.gz"))
  with gzip.open(os.path.join(folder,"uniprot_sprot.fasta.gz"), 'rb') as f_in:
    with open(os.path.join(folder,"uniprot_sprot.fasta"), 'wb') as f_out:
      shutil.copyfileobj(f_in, f_out)
  
  url_uniprot_trembl="https://ftp.expasy.org/databases/uniprot/current_release/knowledgebase/complete/uniprot_trembl.fasta.gz"
  download_file(url_uniprot_trembl,os.path.join(folder,"uniprot_trembl.fasta.gz"))
  with gzip.open(os.path.join(folder,"uniprot_trembl.fasta.gz"), 'rb') as f_in:
    with open(os.path.join(folder,"uniprot_trembl.fasta"), 'wb') as f_out:
      shutil.copyfileobj(f_in, f_out)
  os.remove(os.path.join(folder,"uniprot_trembl.fasta.gz"))
  os.remove(os.path.join(folder,"uniprot_sprot.fasta.gz"))
  os.remove(current_release)
  os.symlink(folder, current_release)

def uniprot_metadata(query_ids,outfile,chunk_size=200):
  def chunks(l, n):
    for i in range(0, len(l), n):
      yield l[i:i + n]  
  headers = { 'User-Agent': 'Mozilla/5.0 (Windows NT 6.3; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko)     Chrome/37.0.2049.0 Safari/537.36'}

  if os.path.isfile(outfile):
    os.remove(outfile)
    
  output=""
  n=0
  for ids in chunks(query_ids,chunk_size):
    n+=1
    m=0
    query= 'id:'+'+OR+'.join(ids)+''+"&format=tab&columns=id,entry name,reviewed,protein names,genes,length,lineage(all),created"
    query_str="https://www.uniprot.org/uniprot/?query={}".format(query)
    print("batch {}".format(n),end="\r")
    response=""
    resp=False
    while not resp:
      try:
        response = requests.get(query_str)
        if response.ok:
          resp=True
        else:
          raise ValueError("not ok")
      except:
        m+=1
        print("batch {}: retry {}".format(n,m),end="\r")
        time.sleep(5)
        pass
    output=response.text 
    with open(outfile,'a') as out:
      out.write(output)
  
def metaclust(folder=None):
  if folder is None:
    try:
      folder=os.environ['METACLUSTFOLDER']
    except:
      raise RuntimeError("The environment variable METACLUSTFOLDER must be defined")
  if not os.path.isdir(folder):
    os.mkdir(folder)
  
  current_release=os.path.join(folder,"current_release")
  today = date.today()
  folder=os.path.join(folder,today.strftime("%Y%m%d"))
  if not os.path.isdir(folder):
    os.mkdir(folder)


  url_metaclust="https://metaclust.mmseqs.org/current_release/metaclust_nr.fasta.gz"
  download_file(url_metaclust,os.path.join(folder,"metaclust-nr.fasta.gz"))
  with gzip.open(os.path.join(folder,"metaclust-nr.fasta.gz"), 'rb') as f_in:
    with open(os.path.join(folder,"metaclust-nr.fasta"), 'wb') as f_out:
      shutil.copyfileobj(f_in, f_out)
  os.remove(os.path.join(folder,"metaclust-nr.fasta.gz"))
  os.symlink(folder, current_release)



#!/usr/bin/env python
import re
import sys

from fastaUtils.grammars import fasta_header_rules

default_int=-1
default_str='None'

def parse(x,type_,default):
  try:
    return type_(x)
  except:
    return default

def iterate_sequences(cmd_beg,cmd_main,cmd_end,sequences,**kwargs):
  for name, value in kwargs.items():
    exec("{} = {}".format(name, value))
  exec(cmd_beg)
  for NR,seq in enumerate(sequences):
    db,uid,name,descr,os,ox,gn,pe,sv,beg,end=parse_header(seq.header.strip())
    exec(cmd_main)
  exec(cmd_end)

class seqRecord:
  def __init__(self,header,sequence=""):
    self.header=header
    self.seq=sequence
  def append(self,sequence):
    self.seq+=sequence
  def __repr__(self):
    return "seqRecord object: {}; sequence length: {}".format(self.header,len(self.seq))
  def __str__(self):
    return ">{}\n{}".format(self.header,self.seq)

def parse_fasta(fastafile):
  s=None
  if fastafile is None:
    for line in sys.stdin:
      if line[0]=='>':
        if s is not None:
          yield s
        s=seqRecord(line.strip()[1:])
      else:
        s.append(line.strip())
    yield s
  else:
    with open(fastafile,'r') as infile:
      for line in infile:
        if line[0]=='>':
          if s is not None:
            yield s
          s=seqRecord(line.strip()[1:])
        else:
          s.append(line.strip())
      yield s

def parse_header(header):
  for r in fasta_header_rules:
    m=re.match(r,header)
    if m is not None:
      d=dict(m.groupdict(default='-'))
      return d.get('db',default_str), d.get('uid',default_str), d.get('name',default_str), d.get('descr',default_str), \
             d.get('os',default_str), parse(d.get('ox',default_int),int,default_int), d.get('gn',default_str), parse(d.get('pe',default_int),int,default_int), \
             parse(d.get('sv',default_int),int,default_int), parse(d.get('beg',default_int),int,default_int), parse(d.get('end',default_int),int,default_int)
  return default_str,default_str,default_str,default_str,default_str,default_int,default_str,default_int,default_int,default_int,default_int

def generate_header(db,uid,name,descr=default_str,os=default_str,ox=default_int,gn=default_str,pe=default_int,sv=default_int,beg=default_int,end=default_int):
  if db!=default_str and name!=default_str:
    if beg!=default_int and end!=default_int:
      header="{}|{}|{}/{}-{}".format(db,uid,name,beg,end)
    else:
      header="{}|{}|{}".format(db,uid,name)
  else:
    if beg!=default_int and end!=default_int:
      header="{}/{}-{}".format(uid,beg,end)
    else:
      header="{}".format(uid)
  if descr!=default_str:
    header+=" "+descr.strip()
  if gn!=default_str:
    header+=" OS={} OX={} GN={} PE={} SV={}".format(os,ox,gn,pe,sv)
  elif os!=default_str:
    header+=" OS={} OX={} PE={} SV={}".format(os,ox,pe,sv)
  return header



from lark import Lark, Transformer,Token
import re

## Fasta related grammars ##
fasta_header1=re.compile(r"(?P<db>\w+)\|(?P<uid>\w+)\|(?P<name>\w+)(?:/(?P<beg>\d+)-(?P<end>\d+))?(?P<descr>.+) OS=(?P<os>.+) OX=(?P<ox>\d+) (?:GN=(?P<gn>.+) )?PE=(?P<pe>\d+) SV=(?P<sv>\d+)") # UNIPROT
fasta_header2=re.compile(r"(?P<db>\w+)\|(?P<uid>\w+)\|(?P<name>\w+)(?:/(?P<beg>\d+)-(?P<end>\d+))?(?P<descr>.+)?")                                                                              # UNIPROT without key-value fields
fasta_header3=re.compile(r"(?P<uid>[^ /]+)(?:/(?P<beg>\d+)-(?P<end>\d+))?(?P<descr>.+)?")                                                                                                       # general match >NAME/beg-end DESCR
fasta_header_rules=[fasta_header1,fasta_header2,fasta_header3]


## AWK related grammars and utils ##
AWK_GRAMMAR="""

beginblock : "\BEGIN"block
mainblock : condition block | condition | block
endblock : "\END"block

program : [beginblock] mainblock [endblock]

condition : /[^{}\\\]+/

code : (/[^{}\\\]+/|block)+
block : "{"(code|block)+"}"

%import common.ESCAPED_STRING
%import common.SIGNED_NUMBER
%import common.WS
%ignore WS
"""

class awkParsing:
  def __init__(self):
    self.codestring={"beginblock":"","mainblock":"","endblock":""}
  def build(self,tree,level=0):
    for subtree in tree.children:
      branch=subtree.data
      if branch=="mainblock":
        data=[c.data for c in subtree.children]
        if "condition" not in data or len(self.visit(subtree.children[data.index("condition")],level=0)[:-1].strip())==0:
          self.codestring[branch]="if True:\n"
        else:
          self.codestring[branch]="if("+self.visit(subtree.children[data.index("condition")],level=0)[:-1]+"):\n"
        if "block" not in data:
          self.codestring[branch]+="  print(seq)\n"
        else:
          self.codestring[branch]+=self.visit(subtree.children[data.index("block")],level=1)
      else:
        self.codestring[branch]=self.visit(subtree,level=-1)
    return self.codestring
    
  def visit(self,tree,level,string=""):
    if type(tree)==Token:
      return "".join(["  "*level+value.strip()+"\n" for value in str(tree).strip().split(';')])
    for subtree in tree.children:
      if type(subtree)==Token:
        string+="".join(["  "*level+value.strip()+"\n" for value in str(subtree).strip().split(';')])
      elif subtree.data=="code":
        string=self.visit(subtree,level,string)
      elif subtree.data=="block":
        if len(string):
          string=string[:-1]+":\n"
        string=self.visit(subtree,level+1,string)
    return string

awk_parser = Lark(AWK_GRAMMAR, start="program")
awk_tree = awkParsing()

##



#!/usr/bin/env python

import random

def generate_random_seq(profiledata):
  return "".join([ random.choices(profiledata.uniq[col],weights=profiledata.prob[col])[0] for col in range(len(profiledata.uniq)) ])

def generate_random_seq_with_template(profiledata,template,n_subs):
  cols=random.sample(range(len(template)),k=n_subs)
  newchars=[ random.choices(profiledata.uniq[col],weights=profiledata.prob[col])[0] for col in cols ]
  template=list(template)
  for col,char in zip(cols,newchars):
    template[col]=char
  return "".join(template)

def shuffle(seqs,columns=None,regions=None):
  if columns is None and regions is None:
    random.shuffle(seqs)
    return seqs

  import numpy as np
  x=np.array([[s for s in seq] for seq in seqs],dtype=str)
  if regions is not None:
    for r in range(len(regions)-1):
      np.random.shuffle(x[:,regions[r]+1:regions[r+1]])
  elif columns is not None:
    for col in columns:
      np.random.shuffle(x[:,col])
  seqs=["".join(x[i,:]) for i in range(x.shape[0])]
  return seqs



from setuptools import setup

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(name='fastaUtils',
      version='1.0.0',
      description='A set of scripts to ease the process of manipulating fasta files',
      long_description=long_description,
      long_description_content_type="text/markdown",
      url='https://gitlab.com/LBS-EPFL/code/fastautils/-/tree/v3.0',
      author='szamuner',
      author_email='stefano.zamuner@protonmail.com',
      scripts=['bin/fst-awk','bin/fst-download','bin/fst-grep','bin/fst-paste','bin/fst-shuf','bin/fst-cut','bin/fst-encode','bin/fst-logo','bin/fst-profile','bin/fst-sort','bin/fst-distance','bin/fst-fromstockholm','bin/fst-random','bin/fst-split'],
      license='MIT',
      packages=['fastaUtils'],
      zip_safe=False,
      classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Development Status :: 4 - Beta",
        "Intended Audience :: Science/Research",
        "Operating System :: POSIX :: Linux",
        "Topic :: Scientific/Engineering :: Bio-Informatics",
      ],
      python_requires='>=3.6',
    )
